package ru.dbotthepony.mc.otm.data

import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.JsonOps
import net.minecraft.data.recipes.FinishedRecipe
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeSerializer
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.util.readBinaryJsonWithCodecIndirect
import ru.dbotthepony.mc.otm.core.util.writeBinaryJsonWithCodec
import kotlin.collections.ArrayDeque
import kotlin.concurrent.getOrSet

class Codec2RecipeSerializer<S : Recipe<*>> private constructor(
	val empty: S?,

	/**
	 * [ThreadLocal] because optimization mods can (and probably should) parallelize recipe deserialization,
	 * since RecipeSerializers are expected to be stateless. [Codec2RecipeSerializer], however, is stateful (threading PoV).
	 * To make it stateless, [ThreadLocal] is used.
	 */
	private val idStack: ThreadLocal<ArrayDeque<ResourceLocation>>,
	codec: (Codec2RecipeSerializer<S>.Context) -> Codec<S>,
) : Codec<S>, RecipeSerializer<S> {
	constructor(empty: S?, codec: (Codec2RecipeSerializer<S>.Context) -> Codec<S>) : this(empty, ThreadLocal(), codec)
	constructor(supplier: (Codec2RecipeSerializer<S>.Context) -> Codec<S>) : this(null, supplier)

	private val codec = codec.invoke(Context())
	private val id: ArrayDeque<ResourceLocation>
		get() = idStack.getOrSet { ArrayDeque() }

	inner class Context() {
		val id: ResourceLocation
			get() = checkNotNull(this@Codec2RecipeSerializer.id.lastOrNull()) { "Not currently deserializing recipe" }

		fun <O : Recipe<*>> wrap(other: Codec2RecipeSerializer<O>): Codec<O> {
			return object : Codec<O> {
				override fun <T : Any> encode(input: O, ops: DynamicOps<T>, prefix: T): DataResult<T> {
					try {
						other.id.addLast(this@Context.id)
						return other.encode(input, ops, prefix)
					} finally {
						other.id.removeLast()
					}
				}

				override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<O, T>> {
					try {
						other.id.addLast(this@Context.id)
						return other.decode(ops, input)
					} finally {
						other.id.removeLast()
					}
				}
			}
		}
	}

	override fun <T : Any> encode(input: S, ops: DynamicOps<T>, prefix: T): DataResult<T> {
		return codec.encode(input, ops, prefix)
	}

	override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<S, T>> {
		return codec.decode(ops, input)
	}

	fun <O : Recipe<*>> xmap(to: (S) -> O, from: (O) -> S): Codec2RecipeSerializer<O> {
		return Codec2RecipeSerializer(empty?.let(to), idStack) { _ ->
			codec.xmap(to, from)
		}
	}

	override fun fromJson(id: ResourceLocation, data: JsonObject): S {
		try {
			this.id.addLast(id)

			return decode(JsonOps.INSTANCE, data).get().map(
				{
					it.first
				},
				{
					empty ?: throw JsonSyntaxException("Failed to deserialize recipe from JSON: ${it.message()}")
				}
			)
		} finally {
			this.id.removeLast()
		}
	}

	override fun fromNetwork(id: ResourceLocation, data: FriendlyByteBuf): S? {
		try {
			this.id.addLast(id)

			return data.readBinaryJsonWithCodecIndirect(this)
				.resultOrPartial { LOGGER.error("Failed to read recipe $id from network: $it") }.orElse(null)
		} finally {
			this.id.removeLast()
		}
	}

	override fun toNetwork(data: FriendlyByteBuf, recipe: S) {
		data.writeBinaryJsonWithCodec(this, recipe)
	}

	fun toFinished(recipe: S): FinishedRecipe {
		return object : FinishedRecipe {
			override fun serializeRecipeData(p_125967_: JsonObject) {
				encode(recipe, JsonOps.INSTANCE, p_125967_).get().map(
					{
						it as JsonObject

						for ((k, v) in it.entrySet()) {
							p_125967_[k] = v
						}
					},
					{
						throw JsonParseException("Failed to serialize recipe: ${it.message()}")
					}
				)
			}

			override fun getId(): ResourceLocation {
				return recipe.id
			}

			override fun getType(): RecipeSerializer<*> {
				return this@Codec2RecipeSerializer
			}

			override fun serializeAdvancement(): JsonObject? {
				return null
			}

			override fun getAdvancementId(): ResourceLocation? {
				return null
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
